/* TIGRE implementation:  LEXER.C */
/* Lexer for graph reducers */

/* (C) Copyright 1989   Philip Koopman Jr.  */
/*  Last update:  3/1/89   */
/*  Test code version -- not for distribution  */


#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include "lexer.h"
#include "reduce.h"

int handle,status   ;  /* file handling variables */
FILE *f0 ;

char eofchar = EOF ;   /* used for assignments to force 1 byte storage */

int CurrentLine ;      /* stores current input line number */
int Overflow ;         /* overflow flag set if lexeme too long */

int LexemeLength ;     /* length of currently building lexeme */
                  /* Lexeme not removed from input stream unless necessary */

int AtEOF ;            /* flag used instead of just pointing at EOF */
                       /* to avoid problems with lexer at end of ID */

#if DEBUG_LEXER
int debugindex ;       /* index for loops for debug printouts */
#endif

/* Double-buffer area for text input allow room for 2 blocks + 1 for spill */
/* Double-buffering chosen over reading input to \n character for better   *
 *   run-time efficiency (admittedly at the cost of making my program a    *
 *   bit more complicated                                                  *
 */
char InputBuffer[MAXLEN+MAXLEN+1] ;

/* Pre-set variables with limit values for input buffer */
char *StartOfBuffer  = InputBuffer;                 /* 1st buffer area */
char *Midpoint       = InputBuffer+MAXLEN;          /* 2nd buffer area */
char *EndOfBuffer    = InputBuffer+MAXLEN+MAXLEN;   /* spill character */

/* Variables for start of current token & next char to be fetched */
char *StartOfToken ;           /* address of 1st char of current token */
char *NextBufferChar ;         /* next unread character in input stream */

int FirstToSecond ;  /* flag indicating second buffer has new data */
int SecondToFirst ;  /* flag indicating first  buffer has new data */


/* --------------------------------------------------------------- *
 *                    INITIALIZATION ROUTINES                      *
 * --------------------------------------------------------------- */

/* intialize input buffers */
/* presume that the longest sequence of input is MAXLEN characters */
/*  all others may be chopped down to small sizes if too long */
void InitInputBuffer()
  {

    /* Initialize buffer pointers to start of buffer */
    StartOfToken   = StartOfBuffer ;
    NextBufferChar = StartOfBuffer ;
    FirstToSecond  = FALSE         ;
    SecondToFirst  = FALSE         ;

    /* Read in first buffer to get the ball rolling */
    status = fread(StartOfBuffer,1,MAXLEN,f0);

    /* set eof sentinel character just beyond buffer read in      *
     * This technique is used so that eof is automatically found  *
     *   during the FSM search.  That way program doesn't have to *
     *   continually check to see if input pointer is past EOF   */
    *(StartOfBuffer+status) = eofchar;
#if DEBUG_LEXER
    printf("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
    printf("\nInitial buffer read in at StartOfBuffer.  ");
    printf("Length = %d.  Contents are:\n",status);
    for (debugindex = 0; debugindex < status; debugindex++)
        printf("%c",*(StartOfBuffer+debugindex));
    printf("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
#endif
  } ;


/* Outside callable driver for all initialization */
void InitLexer(char *filename)
/* Initialize variables used by lexer */
  { f0 = fopen(filename,"r");
    if (f0 == NULL)
      {  printf("\nfailed to open %s\n",filename);
         exit(-1);
      }
    InitInputBuffer();
    CurrentLine = 1 ;    /* First line of input is line 1 */
    AtEOF = FALSE ;      /* Obviously not at EOF at start */
  }

void EndLexer()
  {  fclose(f0);
  }


/* --------------------------------------------------------------- *
 *                    PRIMITIVE SERVICE ROUTINES                   *
 * --------------------------------------------------------------- */

/* Get next input character from buffered input.                        *
 * Returns EOF value if at end of file.  Also maintains pointers into   *
 *   double buffer and automatically manages buffers.                   *
 * Automatically tracks current lexeme length.                          *
 * Sets the Overflow flag if current lexeme is longer than one of the   *
 *   input buffers (in this case, some of token may be lost, but an     *
 *   error token will be produced in any event)                         */
char nextc()
  {
    char thischar;    /* Current character from input stream */

    if ( AtEOF )  return(EOF);   /* return as many EOF's as asked for */

    /* get character and increment pointer */
    thischar = *NextBufferChar ;
    NextBufferChar += 1 ;
    LexemeLength += 1 ;

    if (thischar == eofchar)   AtEOF = TRUE ;

#if DEBUG_LEXER
    /* echo characters as input.  You'll see lots of repeat chars   *
     * as the backtracking takes place and the same char is re-read */
     printf("%c",thischar);
#endif

    /* see if overflowed from first buffer into second */
    if (NextBufferChar == Midpoint )
      { if (! FirstToSecond)    /* and the buffer is the old one */
          { status = fread(Midpoint,1,MAXLEN,f0);
#if DEBUG_LEXER
            printf("\n~~~~~~~~~~~~~~2nd buffer~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
            printf("\nNew buffer read in at Midpoint.  ");
            printf("Length = %d.  Contents are:\n",status);
            for (debugindex = 0; debugindex < status; debugindex++)
                printf("%c",*(Midpoint+debugindex));
            printf("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
#endif
            /* set character just beyond one read to end of file */
            *(Midpoint+status) = eofchar;

            /* set flags to prevent re-reading in case of backtracking */
            FirstToSecond = TRUE ;
            SecondToFirst = FALSE ;

            /* check for lexeme overflow >= MAXLEN characters */
            if ( StartOfToken >= Midpoint )  Overflow = TRUE ;
          }
      }
    /* see if overflowed from second buffer into first */
    if (NextBufferChar == EndOfBuffer )
      { NextBufferChar = StartOfBuffer ;
        if (! SecondToFirst)   /* and the buffer has the old contents */
          { status=fread(StartOfBuffer,1,MAXLEN,f0);
#if DEBUG_LEXER
            printf("\n~~~~~~~~~~~1st buffer~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
            printf("\nNew buffer read in at StartOfBuffer.  ");
            printf("Length = %d.  Contents are:\n",status);
            for (debugindex = 0; debugindex < status; debugindex++)
                printf("%c",*(StartOfBuffer+debugindex));
            printf("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
#endif
            /* set character just beyond one read to end of file */
            *(StartOfBuffer+status) = eofchar;
#if DEBUG_LEXER
            printf("\nNextBufferChar = %d  ",NextBufferChar);
            printf("Next char is %c = %d \n",*NextBufferChar,*NextBufferChar);
#endif
            /* set flags to prevent re-reading in case of backtracking */
            SecondToFirst  = TRUE ;
            FirstToSecond  = FALSE ;
            /* check for lexeme overflow > 80 characters */
            if ( StartOfToken < Midpoint )  Overflow = TRUE ;
          }
      }
    return(thischar);
  };  /* end of nextc */


/* Retract read-in characters.
 * Inputs:   0 = token has been found.  reset retract pointer to current char.
 *         ALL = retract to start of token (place of last retract(0)).
 *           n = retract n characters.  Compiler error if this is too far
 *               or a negative number.
 * Caution: if lexeme attempted was longer than MAXLEN characters, some or
 *          all of it may be overwritten by buffering.  This will not crash
 *          the compiler, and this long a lexeme is an error anyway.
 */
void retract(int numchars)
  {
    switch(numchars)
      { case 0:    /* token found.  no backups requested on this item */
            Overflow = FALSE ;
            LexemeLength = 0 ;
            StartOfToken = NextBufferChar ;   break ;

        case ALL:  /* back up all the way to last retract(0)  */
            Overflow = FALSE ;
            AtEOF = FALSE ;
            LexemeLength = 0 ;
            NextBufferChar = StartOfToken ;   break ;

        default:   /* back up n characters. */
            /* potential for false overflow error -- but not very likely */
            /* in current compiler, max backup is only 1 character! */
            if ( LexemeLength < numchars )
               /* error condition */
               printf("Compiler Error: retract(%d) is too far\n",numchars);
            else
              {  /* do the retraction */
                AtEOF = FALSE ;
                NextBufferChar -= numchars ;
                LexemeLength -= numchars ;
                if ( NextBufferChar < StartOfBuffer )
                    /* wrap-around the buffer limits */
                    NextBufferChar += ( MAXLEN + MAXLEN ) ;
              }
      }
  };  /* end of retract */


/* return TRUE if c is in the set {a..z, A..Z, _ }  */
/* depends on ASCII character set ordering */
int IsLetter(char c)
  {
    return ( ( c >= 'a' && c <= 'z' )  ||
             ( c >= 'A' && c <= 'Z' )  ||
             ( c == 39  )  ||        /* allow ' as a character for S' */
             ( c == '_' )
           ) ;
  } ;

/* return TRUE if c is in the set {0..9}  */
/* depends on ASCII character set ordering */
int IsDigit(char c)
  {
    return ( c >= '0' && c <= '9' ) ;
  } ;

/* return TRUE if c is any printable standard ASCII characer  */
/* plus tab and space -- used for inside quotes               */
/* depends on ASCII character set ordering */
int IsChar(char c)
  {
    return ( ( c == ' '  )  ||
             ( c == '\t' )  ||
             ( c == '!' )   ||
             ( c >= '#' && c <= '~' )   ) ;
  } ;

/* return TRUE if c is an allowable identifier character  */
/*  for 2nd and following id char          */
/* depends on ASCII character set ordering */
int IsIdentChar(char c)
  {
    return ( ( (c != ')') && (c != '(') && (c != '.') )
          && ( (c == '!') || (c >= '#' && c <= '~') )  ) ;
  } ;


/* Test input string for keyword.  This routine is one of the reasons why the
 *    token strings have a \0 placed at the end of them *
 * The returned value is an enum value for keyword tokens, or NOT_A_TOKEN if
 *    no token match
 */
int TestForKeyword(char *s)
  {
    if (strcmp(s,"I")   == 0)  return(DO_I);
    if (strcmp(s,"K")   == 0)  return(DO_K);
    if (strcmp(s,"S")   == 0)  return(DO_S);
    if (strcmp(s,"S'")  == 0)  return(DO_SPRIME);
    if (strcmp(s,"B")   == 0)  return(DO_B);
    if (strcmp(s,"B*")   == 0)  return(DO_BSTAR);
    if (strcmp(s,"C")   == 0)  return(DO_C);
    if (strcmp(s,"C'")   == 0)  return(DO_CPRIME);
    if (strcmp(s,"IF")  == 0)  return(DO_IF);
    if (strcmp(s,"EQ")  == 0)  return(DO_EQ);
    if (strcmp(s,"P")  == 0)  return(DO_P);
    if (strcmp(s,"U")  == 0)  return(DO_U);
    if (strcmp(s,"NOT")  == 0)  return(DO_NOT);
    if (strcmp(s,"ABS")  == 0)  return(DO_ABS);
    if (strcmp(s,"TRUE")  == 0)  return(DO_TRUE);
    if (strcmp(s,"FALSE")  == 0)  return(DO_FALSE);
    if (strcmp(s,"NIL")  == 0)  return(DO_NIL);
    if (strcmp(s,"PRINT")  == 0)  return(DO_PRINT);
    if (strcmp(s,"NL")  == 0)  return(DO_NL);
    if (strcmp(s,"AND")  == 0)  return(DO_AND);
    if (strcmp(s,"OR")  == 0)  return(DO_OR);
    if (strcmp(s,"MOD")  == 0)  return(DO_MOD);
    if (strcmp(s,"CYCLE")  == 0)  return(CYCLE);
    if (strcmp(s,"$1")  == 0)  return(DO_S1);
    if (strcmp(s,"$2")  == 0)  return(DO_S2);
    if (strcmp(s,"$3")  == 0)  return(DO_S3);
    if (strcmp(s,"$4")  == 0)  return(DO_S4);
    if (strcmp(s,"$5")  == 0)  return(DO_S5);
    if (strcmp(s,"$6")  == 0)  return(DO_S6);
    if (strcmp(s,"$7")  == 0)  return(DO_S7);
    if (strcmp(s,"$8")  == 0)  return(DO_S8);
    if (strcmp(s,"FLOOR")  == 0)  return(DO_FLOOR);
    if (strcmp(s,"$10")  == 0)  return(DO_S10);
    if (strcmp(s,"$11")  == 0)  return(DO_S11);
    if (strcmp(s,"$12")  == 0)  return(DO_S12);
    if (strcmp(s,"$13")  == 0)  return(DO_S13);
    if (strcmp(s,"$14")  == 0)  return(DO_S14);
    if (strcmp(s,"$15")  == 0)  return(DO_S15);
    if (strcmp(s,"$16")  == 0)  return(DO_S16);
    if (strcmp(s,"$17")  == 0)  return(DO_S17);
    if (strcmp(s,"$18")  == 0)  return(DO_S18);
    if (strcmp(s,"$19")  == 0)  return(DO_S19);
    /* not a keyword */
    printf("** illegal operator '%s' **",s);
    error_flag = TRUE ;
    return (NOT_A_TOKEN);
  };


/*  Backtracking word for FSM ensemble.  If the current State
 *  was not successful, backtrack and try the next FSM
 *  Usage:    State = NextTry(State);
 */
int NextTry(int State)
  {  /* return next FSM starting state given current state */
     /* use ranges instead of cases to be safer in case of modifications */
    retract(ALL);   /* backtrack input */
    if (State < 20)  return(20);
    if (State < 40)  return(40);
    if (State < 50)  return(50);
    if (State < 60)  return(60);
    if (State < 70)  return(70);
    return(99);
  } ;  /*  end of NextTry */


/* SetToken takes the token address and desired token type as inputs,
 *   sets the token type, and does all the dirty work needed just before
 *   returning with a token.
 * Specifically, it:
 *   sets the token type per the input
 *   if the lexeme is too long (Overflow) it forces token type to error
 *   if the lexeme should be a string, it extracts the string from the
 *     input buffer and sets the string length and pads the string with
 *     a \0 for good behavior  (the \0 is not included in the length)
 *   performs a retract(0) to indicate the token was accepted
 */
void SetToken(token *tok,int TokenType)
  { /* do all the dirty work before exiting with token */
    int i ;
    char *charptr ;

    /* set token type */
    tok->toktype = TokenType ;

    /* if overflowed lexeme, set error flag */
    if (Overflow)
      {  tok->toktype = NOT_A_TOKEN ;
         printf("\nBuffer overflow error");
         error_flag = TRUE ;
      }

    /* copy string to token if required */
    /* pad string with \0 to ensure a valid C string */
     if(TokenType == KEYWORD)
     { if (LexemeLength >= MAXLEN )
          {  /* too long a lexeme */
            printf("\nError - Lexeme too long");
            tok->toktype = NOT_A_TOKEN ;
            error_flag = TRUE ;
            LexemeLength = MAXLEN - 1 ;
          }
         tok->toklen = LexemeLength ;
         charptr = StartOfToken ;

         for (i = 0; i < LexemeLength; i++)
           { /* copy string */
             tok->toktext[i] = *charptr++ ;
             /* wrap around end of buffer if required */
             if (charptr >= EndOfBuffer ) charptr = StartOfBuffer ;
           }
         /* pad lexeme with null so it's a valid string */
         tok->toktext[LexemeLength] = '\0' ;
     }
    /* do a retract(0) to gobble text from input stream */
    retract(0);
  } ;


/* --------------------------------------------------------------- *
 *                  MAIN FSM DRIVING ROUTINE                       *
 * --------------------------------------------------------------- */

/* GetNextToken returns the appropriate token type, line number and
 *   lexeme operands as required by the assignment
 * The input is a pointer to the token data structure supplied by
 *   the calling routine
 */
void GetNextToken(token *tok)
  {
  char InputChar;           /* current input character */
  int State;                /* current state for FSM   */

  int IntConst ;            /* integer constant accumulator */
  int IntSign ;             /* +1 or -1 for sign of IntConst */
  /* float FloatConst ;        /* float   constant accumulator */
  /* float FloatMult  ;        /* multiplier for fractional float part */

  State = 0 ;   /* Initialize for each token */
  Overflow = FALSE ;
  LexemeLength = 0 ;

    /* set token line number */
    tok->linenum = CurrentLine ;

  while(TRUE)     /* loop until return executed for found token */
   {
    /* attempt next FSM transition */
    switch (State)
    {
      /*   0 = whitespace scanning state                                */
      case 0:
          if ( (InputChar = nextc()) == '\n')  /* new input line */
            { CurrentLine += 1 ;
              tok->linenum = CurrentLine ;
              retract(0);
            }
          else
              if ( InputChar == ' ' || InputChar == '\t' )
                  /* discard the whitespace */
                  retract(0);
              else   /* non-whitespace */
                  State = NextTry(State);
          break;


      case 20:
      /*   20 = test/accept for symbols                     */
          if ( (InputChar = nextc()) == '(')
            { SetToken(tok,LPAREN);
              return;
            }
          if (InputChar == ')')
            { SetToken(tok,RPAREN);
              return;  }
          if (InputChar == '+')
            { SetToken(tok,DO_PLUS);
              return;  }
          if (InputChar == '*')
            { SetToken(tok,DO_TIMES);
              return;  }
          if (InputChar == '/')
            { SetToken(tok,DO_DIV);
              return;  }
          if (InputChar == '<')
            { SetToken(tok,DO_LESS);
              return;  }
          if (InputChar == '>')
            { SetToken(tok,DO_GREAT);
              return;  }
         if (InputChar == '=')
            { SetToken(tok,DO_EQ);
              return;  }
      State = NextTry(State);
      break;

      /*   40 = start of node #FSM  */
      /*       41 = #       42 = Integer accepting state                */
      case 40:
          InputChar = nextc();
          if ( InputChar == '#' )
            {  IntConst = 0 ;
               State = 41;
            }
          else  State = NextTry(State);
          break ;

      case 41:
          InputChar = nextc();
          if ( IsDigit(InputChar) )
              IntConst = (IntConst * 10) + (InputChar - '0') ;
          else
              State = 42 ;     /* accept int */
          break ;

      case 42:
          retract(1);   /* un-do the non-integer char */
          SetToken(tok,DO_NODE);
          tok->value = IntConst ;
          return;


      /*   50 = start of identifier/keyword FSM                         */
      /*       51 = Alpha   52 = AlphaNumeric    53 = accept            */
      case 50:
          InputChar = nextc();
          if ( IsLetter(InputChar) || InputChar == '$' )
            { State = 51;
            }
          else  State = NextTry(State);
          break ;

      case 51:
          InputChar = nextc();
          if ( IsIdentChar(InputChar) )
            { State = 51;
            }
          else  State = 52 ;
          break ;

      case 52:
          retract(1);   /* un-do the non-identifier char  */
          SetToken(tok,KEYWORD);
          tok->toktype = TestForKeyword(tok->toktext) ;
          return;


      /*  60 = minus (possibly for a number) */
      case 60:
          InputChar = nextc();
          if (InputChar == '-')
            {  IntConst = 0 ;
               IntSign = -1 ;
               State = 61;
            }
          else  State = NextTry(State);
          break ;

       /* 61 = either unary minus or a subtraction */
      case 61:
          InputChar = nextc();
          if ( IsDigit(InputChar) )
            {  IntConst = InputChar - '0' ;
               State = 71;
            }
          else
            { retract(1);  /* found a subtraction */
              SetToken(tok,DO_MINUS);
              return;  }
          break ;

      /*   70 = start of constant FSM    ( note: no unary '-' in spec!) */
      /*       71 = Digit   72 = Integer accepting state                */
      case 70:
          InputChar = nextc();
          if ( IsDigit(InputChar) )
            {  IntConst = InputChar - '0' ;
               IntSign = 1 ;
               State = 71;
            }
          else  State = NextTry(State);
          break ;

      case 71:
          InputChar = nextc();
          /* overflow of integer not checked, but I thought of it! */
          /*         this gets tricky if machine word sizes change */
          /* for now, return result mod wordsize */
          if ( IsDigit(InputChar) )
              IntConst = (IntConst * 10) + (InputChar - '0') ;
          else
              State = 72 ;     /* accept int */
          break ;

      case 72:
          retract(1);   /* un-do the non-integer char */
          SetToken(tok,DO_LIT);
          if (IntSign < 0)  IntConst = 0 - IntConst ;
          tok->value = IntConst ;
          if (IntConst == 1)
            { tok->toktype = DO_ONE ;
              return;  }
          if (IntConst == 0)
            {  tok->toktype = DO_ZERO ;
              return;  }
          if (IntConst == 2)
            {  tok->toktype = DO_TWO ;
              return;  }
          if (IntConst == 3)
            {  tok->toktype = DO_THREE ;
              return;  }
          return;

      /* States for FSM -- Uses backtracking */
      /*  99 = end of file check & read in character  */
      case 99:
          if ( (InputChar == eofchar) || (InputChar == '.') )
            { SetToken(tok,EOF_TOKEN) ;
            }
          else   /* bad character found -- no match */
            {
#if DEBUG_LEXER
              printf("\nBad character in input stream = %c",InputChar);
              printf("  Value= %d\n",InputChar);
#endif
              error_flag = TRUE ;
              SetToken(tok,NOT_A_TOKEN);
            }  ;
         return;

      /* default case is an error condition */
      default:   /* bad case number */
         printf("Compiler error in main switch statement.. case=%d\n",State) ;
         printf("current character is %c\n",InputChar);
         State = 0 ;
         retract(0);

    }   /* case */
   }    /* while forever */
  } ;   /* GetNextToken */
